/***************************************************************************
 *   Copyright (C) 2005 by Timothy Redaelli <drizzt@freesbie.org>          *
 *   Copyright (C) 2004 vu1tur <v@vu1tur.eu.org>                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "base64.h"

inline bool is_base64(const char c) {
	if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c == '+' || c == '/' || c == '=')
		return true;
	return false;
}

void cleanup_base64(char *inp, const uint32_t size) {
	char *tinp1, *tinp2;
	uint32_t i;
	tinp1 = inp;
	tinp2 = inp;

	for (i = 0; i < size; i++) 
		if (is_base64(*tinp2)) 
			*tinp1++ = *tinp2++;
		else 
			*tinp1 = *tinp2++;
	*(tinp1) = 0;
}

unsigned char decode_base64_char(const char c) {
	if (c >= 'A' && c <= 'Z')
		return c - 'A';
	if (c >= 'a' && c <= 'z')
		return c - 'a' + 26;
	if (c >= '0' && c <= '9')
		return c - '0' + 52;
	if (c == '+')
		return 62;
	if (c == '=')
		return 0;
	return 63;
}

bool decode_base64(const char *inp, uint32_t isize, char *out, uint32_t *osize) {
	char *tinp = (char*)inp;
	char *tout;
	uint32_t i;

	*osize = isize / 4 * 3;
	if (inp != out)
		out = tout = (char*) malloc(*osize);
	else
		tout = tinp;

	for(i = 0; i < isize >> 2; i++) {
		*tout++ = (decode_base64_char(*tinp++) << 2) | (decode_base64_char(*tinp) >> 4);
		*tout++ = (decode_base64_char(*tinp++) << 4) | (decode_base64_char(*tinp) >> 2);
		*tout++ = (decode_base64_char(*tinp++) << 6) | decode_base64_char(*tinp++);
	}
	if (*(tinp-1) == '=')
		(*osize)--;
	if (*(tinp-2) == '=')
		(*osize)--;
}
