/***************************************************************************
 *   Copyright (C) 2005 by Timothy Redaelli <drizzt@freesbie.org>          *
 *   Copyright (C) 2004 vu1tur <v@vu1tur.eu.org>                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "dmg2iso.h"

uint32_t findstr(const char* inp, const char* str, const uint32_t ilen, const uint32_t slen, const uint32_t num) {
	uint32_t tnum;
	int found = 0;
	char *tinp, *tstr;
	if(num > 0) {
		tnum = num - 1;
		for(tinp = (char *)inp; tinp < inp + ilen - slen + 1; tinp++) {
			for(tstr = (char *)str; tstr < str + slen; tstr++) {
				if(*tstr != tinp[tstr - str])
					break;
				if((*tstr == tinp[tstr - str]) && (tstr - str == slen - 1))
					found = 1;
			}
			if((found == 1) && (!tnum--))
				return tinp - inp;
			found = 0;
		}
		return -1;
	} else {
		tnum = 0;
		for(tinp = (char *)inp; tinp < inp+ilen-slen+1; tinp++) {
			for(tstr = (char *)str; tstr < str + slen; tstr++) {
				if(*tstr != tinp[tstr - str])
					break;
				if((*tstr == tinp[tstr - str]) && (tstr - str == slen - 1))
					tnum++;
			}
		}
		return tnum;
	}
}

inline uint32_t convert_endian(uint32_t n) {
	return (n >> 24) | ((n & 0x00ff0000) >> 8) | ((n & 0x0000ff00) << 8) | (n << 24);
}

int main(int argc, char *argv[])
{
	char *ar, *plist, **parts;
	long ss;
	FILE *FIN, *FOUT;
	uint32_t pl_size, last_offs, *partlen;
	long ploffs_begin;
	uint8_t partnum;
	int i;
	Bytef *tmp, *otmp;
	z_stream z;

	puts("dmg2iso 0.1 by Timothy Redaelli <drizzt@gufi.org>");

	if(argc != 3)  {
		printf("Syntax: dmg2iso <filename.dmg> <filename.iso>\n");
		return EXIT_FAILURE;
	}

	FIN = fopen(argv[1], "rb");
	if(FIN == NULL) {
		perror("Can't open input file");
		return EXIT_FAILURE;
	}

	fseek(FIN, -PLIST_LOOKUP_SIZE, SEEK_END);
	ss = ftell(FIN);

	printf("reading property list...");
	ar = (char*)malloc(PLIST_LOOKUP_SIZE);
	if(ar == NULL) {
		perror("\n");
		return EXIT_FAILURE;
	}
	fread((char *)ar, PLIST_LOOKUP_SIZE, 1, FIN);
	if(findstr(ar, PLIST_BEGIN, PLIST_LOOKUP_SIZE, strlen(PLIST_BEGIN), 0) == 0) {
		fputs("\nERROR: Property list is corrupted.\n", stderr);
		return EXIT_FAILURE;
	}
	puts(" ok");

	FOUT = fopen(argv[2], "wb");
	if(FOUT == NULL) {
		perror("Can't open output file");
		return EXIT_FAILURE;
	}

	ploffs_begin = findstr(ar, PARTLIST_BEGIN, PLIST_LOOKUP_SIZE, strlen(PARTLIST_BEGIN), 1);
	pl_size = findstr(ar, PARTLIST_END, PLIST_LOOKUP_SIZE, strlen(PARTLIST_END), 1) + 8 - ploffs_begin;
	plist = (char*)malloc(pl_size);
	memcpy(plist, ar + ploffs_begin, pl_size);
	partnum = findstr(plist, PART_BEGIN, pl_size, strlen(PART_BEGIN), 0);
	printf("found %d partition%s\n", partnum, partnum != 1 ? "s" : "");

	*parts = (char *)malloc(partnum * sizeof(uint8_t));
	partlen = (uint32_t *)malloc(partnum * sizeof(uint32_t));
	
	for (i = 0; i < partnum; i++) {
		uint32_t data_begin = findstr(plist, PART_BEGIN, pl_size,strlen(PART_BEGIN), i + 1) + strlen(PART_BEGIN);
		uint32_t data_size = findstr(plist, PART_END, pl_size, strlen(PART_END), i + 1) - data_begin;
		parts[i] = (char*)malloc(data_size);
		memcpy(parts[i], plist + data_begin, data_size);
		cleanup_base64(parts[i], data_size);
		decode_base64(parts[i], strlen(parts[i]), parts[i], &partlen[i]);
	}
	
	z.zalloc = (alloc_func)0;
	z.zfree = (free_func)0;
	z.opaque = (voidpf)0;
	puts("decompressing.");
	tmp = (Bytef*)malloc(0x40000);
	otmp = (Bytef*)malloc(0x40000);
	last_offs = 0;
	for(i = 0; i < partnum; i++) {
		uint8_t offset = 0xCC;
		uint32_t out_offs, out_size, in_offs, in_size, last_in_offs;
		enum BlockType block_type = 0;
		printf("partition %d...", i);
		while(block_type != 0xffffffff) {
			memcpy(&block_type, parts[i] + offset, 4);
			memcpy(&out_offs, parts[i] + offset + 12, 4);
			memcpy(&out_size, parts[i]+ offset + 20, 4);
			memcpy(&in_offs, parts[i] + offset + 28, 4);
			memcpy(&in_size, parts[i] + offset + 36, 4);
			block_type = convert_endian(block_type);
			out_offs = convert_endian(out_offs) * 0x200;
			out_size = convert_endian(out_size) * 0x200;
			in_offs = convert_endian(in_offs);
			in_size = convert_endian(in_size);
			switch(block_type) {
				char *zeroblock;
				case BT_ZLIB:		/* Decompress */
					if(inflateInit(&z) != Z_OK) {
						fputs("ERROR: Can't initialize inflate stream\n", stderr);
						return EXIT_FAILURE;
					}
					fseek(FIN, last_offs + in_offs, SEEK_SET);
					fread(tmp, in_size, 1, FIN);
					z.next_in = (Bytef*)tmp;
					z.next_out = (Bytef*)otmp;
					while(1) {
						int err = inflate(&z, Z_NO_FLUSH);
						z.avail_in = z.avail_out = 32768;
						if (err != Z_OK && err != Z_STREAM_END) {
							fputs("ERROR: Inflation failed\n", stderr);
							return EXIT_FAILURE;
						}
						if(err == Z_STREAM_END)
							break;
					}
					if (inflateEnd(&z) != Z_OK) {
						fputs("ERROR", stderr);
						return EXIT_FAILURE;
					}
					fwrite(otmp, out_size, 1, FOUT);
					last_in_offs = in_offs + in_size;
					break;
				case BT_COPY:		/* Copy */
					fseek(FIN, last_offs + in_offs, 0);
					fread(tmp, in_size, 1, FIN);
					fwrite(tmp, out_size, 1, FOUT);
					break;
				case BT_ZERO:
					zeroblock = (char*)malloc(out_size);
					memset(zeroblock, 0, out_size);
					fwrite(zeroblock, out_size, 1, FOUT);
					free(zeroblock);
					break;
				case BT_END:
					last_offs += last_in_offs;
					break;
			}
			offset += 0x28;
		}
		puts("ok");
	}
	puts("conversion successful!");
	free(tmp);
	free(otmp);
	for (i = 0; i < partnum; i++)
		free(parts[i]);
	free(plist);
	free(ar);
	fclose(FIN);
	fclose(FOUT);
	return EXIT_SUCCESS;
}
